/*
 * OpenEAS Experlog Application Server
 * Web application development and dynamic template scripting engine
 * Copyright (c) 2004 ExperLog SARL - http://www.experlog.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version. This program is distributed in
 * the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
 * the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have
 * received a copy of the GNU General Public License along with this program;
 * if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package org.experlog.actions;

import org.experlog.util.Client;
import org.experlog.base.ESServletRequest;
import org.experlog.base.ESAction;
import org.experlog.gencode.DynHtml;

/**
 * Generate a message digest (using MD5 or other algorithms)
 * Input parameters:
 *  MDMessage (mandatory): the message
 *  MDAlgorithm (optional): the algorithm - default MD5
 * Output parameters:
 *  MDDigest: The message digest
 */
public class MessageDigest implements ESAction {
    
  public boolean serverSide() { return true; }
    
  public boolean processRequest(Client session, ESServletRequest req,
  DynHtml page, Object info) throws Exception {

    String msg = req.getParameter("MDMessage");
    if(msg == null) {
      req.addParameter("MissingField", "MDMessage");
      return false;
    }
    String algorithm = req.getParameter("MDAlgorithm");
    if(algorithm == null) algorithm = "MD5";

    java.security.MessageDigest md = java.security.MessageDigest.getInstance(algorithm);
    md.update(msg.getBytes());
    req.addParameter("MDDigest", hexString(md.digest()));

    return true;
  }

  private static final char[] HEX_CHARS = new char[]
   {'0','1','2','3','4','5','6','7',
    '8','9','a','b','c','d','e','f'};

  private String hexString (byte b[]) {
   StringBuffer res = new StringBuffer();
   for(int i=0; i<b.length; i++) {
     res.append( HEX_CHARS [ ( b[i]  & 0xf0 ) >>> 4  ] ) ;
     res.append( HEX_CHARS [ b[i] & 0x0f ] ) ;
   }
   return res.toString();
  }

  public static void main(String args[]) throws Exception {
    if(args.length < 1) {
      System.err.println("Usage: org.experlog.actions.MessageDigest message [algorithm]");
      System.exit(1);
    }
    String algorithm = (args.length > 1 ? args[1] : "MD5");
    java.security.MessageDigest md = java.security.MessageDigest.getInstance(algorithm);
    md.update(args[0].getBytes());
    System.out.println(algorithm + " digest for message is: [" + new MessageDigest().hexString(md.digest()) + "]");
  }
}

