/*
 * OpenEAS Experlog Application Server
 * Web application development and dynamic template scripting engine
 * Copyright (c) 2004 ExperLog SARL - http://www.experlog.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version. This program is distributed in
 * the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
 * the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have
 * received a copy of the GNU General Public License along with this program;
 * if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Portions added by jmo are Copyright (c) OSAMES 2009
 */

package org.experlog.gencode;

import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
//import java.io.RandomAccessFile;
import java.io.File;

import java.lang.reflect.Field;
import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;
import java.sql.Statement;
import java.text.DateFormat;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.text.ParsePosition;
import java.text.SimpleDateFormat;
import java.util.Enumeration;                        
import java.util.Hashtable;
import java.util.Locale;
import java.util.StringTokenizer;
import java.util.Vector;

import org.apache.log4j.Category;
import org.experlog.base.ESAction;
import org.experlog.base.ESServletRequest;
import org.experlog.base.OpenClient;
import org.experlog.db.DatasourceProperty;
import org.experlog.db.ESCallable;
import org.experlog.db.ESDBDataSet;
import org.experlog.openeas.api.DataSet;
import org.experlog.openeas.api.DataObject;
import org.experlog.util.BufferedRandomAccessFile;
import org.experlog.util.Client;
import org.experlog.util.ESCategory;
import org.experlog.util.ESEnumDataSet;
import org.experlog.util.ESFileDataSet;
import org.experlog.util.ESMaterializeDataSet;
import org.experlog.util.StringUtils;


class ESLoopExitException extends Exception {
	private static final long serialVersionUID = -6229666497501009833L;
	public ESLoopExitException() {
        super();
    }
    public ESLoopExitException(String msg) {
        super(msg);
    }
};

class CmdTokenizer {

  private static String[] tokenize1(String cmd) {
    char str[] = new char[cmd.length()];
    cmd.getChars(0, cmd.length(), str, 0);
    boolean quoted = false;
    for(int i=0; i<str.length; i++) {
      if(Character.isWhitespace(str[i])) {
        if(quoted) str[i] = '\u0000';
        else str[i] = ' ';
      } else if(str[i] == '\'' || str[i] == '\"') {
        quoted = ! quoted;
        str[i] = '\"';
      }
    }
    return (new String(str)).split("[\"]");
  }
  public static String[] tokenize(String cmd) {
    String s[] = tokenize1(cmd);
    Vector v = new Vector();
    for(int i=0; i<s.length; i++) {
      StringTokenizer st = new StringTokenizer(s[i]);
      while(st.hasMoreTokens()) v.addElement(st.nextToken());
    }
    String ret[] = new String[v.size()];
    for(int i=0; i<v.size(); i++) {
      ret[i] = ((String)v.elementAt(i)).replace('\u0000',' ');
    }
    return ret;
  }
}

/**
 * A Dynamic HTML expander.
 */
public class ESDynHtml implements DynHtml {

    File input_ = null;
    int lnb_ = 0;
    boolean inner_ = false; // true when "inner" template (eg. ExpandFile / Include)
    OpenClient session_ = null;
    ESServletRequest req_ = null;
    boolean hasResults_ = false;
    boolean errorSet_ = false;
    boolean displayUnresolved_ = false;

    int looplevel_ = 0;
    int iflevel_ = 0;

    Hashtable requests_ = new Hashtable();
    Hashtable params_ = null;
    Hashtable loopstart_ = new Hashtable();
    Hashtable datasets_ = new Hashtable();
    Hashtable dataobjects_ = new Hashtable();
    Hashtable customStatements_ = null;
    
    // stored procedures
    Hashtable procedures_ = new Hashtable();

    Locale formatLocale_ = Locale.getDefault(); // Locale for number/date formats

    ESDateAccess dateaccess_;
    Category logger_ = null;

    // created objects by user (UserObject)
    Hashtable uObjects_ = null;
    String encodingValue = null;

    public ESDynHtml() { }

    /**
     * Create a new DynHtml expander
     * @param session A JDBC Connection (can be null)
     * @param file An input file (can be null)
     */
    public ESDynHtml(Client session, File file) {
        this(session, file, true);
    }

    /**
     * Create a new DynHtml expander
     * @param session A JDBC Connection (can be null)
     * @param file An input file (can be null)
     * @param enforceSecurity If true, enforce security check on file name
     */
    public ESDynHtml(Client session, File file, boolean enforceSecurity) {
      try {
        session_ = (OpenClient) session;
        setInputFile(file, enforceSecurity);

        if (session != null) {
          logger_ =
            ESCategory.getCategory(
                session.getAppName(),
                "openeas.gencode");
          String unr =
            session.getAppConfiguration().get(
                "displayUnresolvedNames");
          if (unr != null && unr.trim().toLowerCase().equals("true"))
            displayUnresolved_ = true;

          formatLocale_ = session.getAppConfiguration().getFormatLocale();
        }

        dateaccess_ = new ESDateAccess();
        addDataObject("CurrentDate", dateaccess_);
        
        addDataObject("ResourceBundle", new ResourceBundleAccess(this));

        encodingValue = session.getAppConfiguration().getEncodingValue();

        // No null logger allowed
        if (logger_ == null)
          logger_ = ESCategory.getCategory("openeas.gencode");

      } catch (Exception e) {

        e.printStackTrace(System.err);
        logError("Internal error: " + e);
      }
    } // constructor

    public boolean isInner() { return inner_; }
    public void setInner(boolean inner) { inner_ = inner; }
    
    public OpenClient getSession() { return session_; }
    
    void logError(String info) {
      if (logger_ == null)
        logger_ = ESCategory.getCategory("openeas.gencode");
      String msg = "["
                + new java.util.Date(System.currentTimeMillis())
                + "] ERROR IN TEMPLATE "
                + input_.getName()
                + " line " + lnb_ + ": "
                + info;
      logger_.error(msg);
      System.err.println(msg);
    }

    String getPositionInfo(boolean timestamp) {
      return (timestamp ?
       "[" + new java.util.Date(System.currentTimeMillis()) + "] " : "")
       + "[ TEMPLATE " + input_.getName() + " line " + lnb_ + " ] ";
    }

    /**
     * Display unresolved symbols or not.
     */
    public void displayUnresolved(boolean yes) {
        displayUnresolved_ = yes;
    } // method displayUnresolved

    /**
     * Change the input file definition
     * @param file An input file
     */
    public void setInputFile(File file) throws IOException {
        setInputFile(file, true);
    }

    /**
     * Change the input file definition
     * @param file An input file
     * @param enforceSecurity If true, enforce security check on file name
     */
    public void setInputFile(File file, boolean enforceSecurity)
        throws IOException {
        if (file != null
            && enforceSecurity
            && !org.experlog.util.AppConfiguration.isSecurePath(file.getPath()))
            throw new IOException(
                "ExpandFile: Access denied to " + file.getPath());
        input_ = file;
    } // method setInputFile

    /**
     * Define a data set
     * @param ds An ESDataSet object
     */
    public void addDataSet(String name, DataSet ds) {
      DataSet d = (DataSet) datasets_.remove(name);
      if (d != null) {
        try {
          d.close();
        } catch (Exception ignore) { }
      }
      datasets_.put(name, ds);
    } // method addDataSet

    /** public ESDynHtml(Client session, File file, boolean enforceSecurity) {
      super(session, file, enforceSecurity);
    
     * Define a data set
     * @param t An ESDataSet object
     */
    public void addDataObject(String name, DataObject t) {
        dataobjects_.put(name, t);
    } // method addDataObject

    /**
     * Define a new parameter (a tagname/value association)
     * When $tagname$ will be found, it will be expanded to value.
     */
    public void setParameter(String name, String val, boolean addtoreq) {
        if (params_ == null)
            params_ = new Hashtable();

        params_.put(name, val);

        if (addtoreq && req_ != null)
            req_.addParameter(name, val);
    } // method setParameter

    public void setParameters(Hashtable h, boolean addtoreq) {
        if (params_ == null)
            params_ = h;
        else {
            Enumeration en = h.keys();
            while (en.hasMoreElements()) {
                Object key = en.nextElement();
                Object val = h.get(key);
                params_.put(key, val);
                if (addtoreq && req_ != null)
                    req_.addParameter(key.toString(), val.toString());
            }
        }
    } // method setParameters

    public void setServletRequest(ESServletRequest req) {
        req_ = req;
        setParameters(req.getAllParams(), false);
        addDataObject(
            "HttpHeaders",
            new ESHttpHeadersAccess(req.getServletRequest()));
    } // method setServletRequest

    /**
     * Close any data set known
     * This prevents database cursors to remain open!
     */
    private void closeDataSets() {
        Enumeration nm = datasets_.keys();

        while (nm.hasMoreElements()) {
            DataSet ds = (DataSet) datasets_.get(nm.nextElement());
            try {
		ds.close();
            } catch (Exception e) {
            }
        }
    }

    /**
     * Close any procedures known
     */
    private void closeProcedures() {
        Enumeration nm = procedures_.keys();

        while (nm.hasMoreElements()) {
            ESCallable ds = (ESCallable) procedures_.get(nm.nextElement());
            try {
                ds.close();
            } catch (Exception e) {
                logger_.error("Problem to close procedure");
            }
        }
    }

    private void closeUserObjects() {
        if (this.uObjects_ != null) {
            Enumeration nm = this.uObjects_.keys();

            while (nm.hasMoreElements()) {
                UserObject uo = (UserObject) uObjects_.get(nm.nextElement());
                uo.close();
                uo = null;
            }
        }
    }

    /**
     * Expand the HTML template on a given output stream.
     */
    public void expand(PrintWriter out) throws ExitException {
        expand(out, true);
        setPriority(Thread.NORM_PRIORITY);
    }

    /**
     * Expand the HTML template on a given output stream.
     */
    public void expand(PrintWriter out, boolean closeds) throws ExitException {
        BufferedRandomAccessFile in = null;
        try {
            in = new BufferedRandomAccessFile(input_);
            String line;
            while ((line = in.readLine()) != null) {
                lnb_++;
                // Modified by jmo for OSAMES Team 2009
                byte[] utf8BytesLine = line.getBytes("ISO-8859-1");
                expandLine(new String(utf8BytesLine, "UTF-8"), in, out);
                // End jmo modifications
                //expandLine(line, in, out);
            }
            flush(out);
            in.close();
        } catch (ExitException e) {
            flush(out);
            if (in != null) try { in.close(); } catch (IOException ignore) { }
            if(e.exitAll() && inner_) throw e;
        } catch (Exception e) {
          	logError(e.toString());

            e.printStackTrace(System.err);

            out.println("<p><b>["
             + new java.util.Date(System.currentTimeMillis())
             + "] INTERNAL ERROR IN " + input_.getName()
             + " line " + lnb_ + ": "
             + e + "</b></p>");
            flush(out);
            if (in != null)
                try {
                    in.close();
                } catch (IOException ioe) {
                }
        } finally {
            if (closeds) {
                closeDataSets(); // make sure all known data sets are closed!
                closeProcedures();
                closeUserObjects();
            }
        }
    } // method expand

    void expandLine(String line, BufferedRandomAccessFile in, PrintWriter out)
        throws Exception {

        if(startsWW(line, "$EXITALL"))
          throw new ExitException(true);
        else if (startsWW(line, "$EXIT"))
          throw new ExitException(false);

        if (isComment(line))
            return;

        CustomStatement stmt = null;
        
        if (isAdmOnly(line)) {
            checkAdmLogin();
        } else if (isDebug(line)) {
            traceLine(line);
        } else if (isAction(line)) {
            runAction(line.trim());
        } else if (isDefineSql(line)) {
            defineSql(line.trim());
        } else if (isAppendSql(line)) {
            appendSql(line.trim());
        } else if (isExecSql(line)) {
            execSql(line.trim());
        } else if(isExpandSql(line)) {
          expandSql(line.trim());
        } else if (isLoopBegin(line)) {
            expandLoop(line.trim(), in, out);
        } else if (isExpandFile(line)) {
            expandFile(line.trim(), out, false);
        } else if (isInclude(line)) {
            expandFile(line.trim(), out, true);
        } else if (isCookie(line)) {
            handleCookie(out);
        } else if (isAssign(line) || isDefVar(line)) {
            assignParam(line.trim());
        } else if (isSetCookie(line)) {
            setCookie(line.trim());
        } else if (isAppend(line)) {
            appendParam(line.trim());
        } else if (isCalc(line)) {
            doCalc(line.trim());
        } else if (isCheckNumber(line)) {
            checkNumber(line.trim());
        } else if (isClearError(line)) {
            errorSet_ = false; // clear error
        } else if (isIf(line)) {
            expandIf(line, in, out, true);
        } else if (isMaterialize(line)) {
            expandMaterialize(line.trim());
        } else if (isSystem(line)) {
            expandSystem(line.trim(), out);
        } /*else if(isEval(line)) {
        			expandEval(line, in, out, params_);
        		}*/
        else if (isSwitch(line)) {
            expandSwitch(line, in, out, true);
        } else if (isDefineSqlProc(line)) {
            expandDefineSqlProc(line.trim());
        } else if (isRegisterOutParameter(line)) {
            expandRegisterOutParameter(line.trim());
        } else if (isExecSqlProc(line)) {
            expandExecSqlProc(line.trim());
        } else if (isNewObject(line)) {
            expandNewObject(line.trim());
        } else if (isInvoke(line)) {
            expandInvoke(line.trim());
        } else if (isCloseObject(line)) {
            expandCloseObject(line.trim());
        } else if(isForward(line)) {
          expandForward(line);
        } else if((stmt = getStatement(line)) != null) {
          stmt.execute(line.trim(), in, out, this);
        } else {
            println(out, expandString(line, params_));
        }
    } // method expandLine

    public void addStatement(String name, CustomStatement stmt) {
      if(customStatements_ == null) customStatements_ = new Hashtable();
      customStatements_.put(name.toLowerCase(), stmt);
    }
    public CustomStatement getStatement(String line) {
      if(customStatements_ == null) return null;
      StringTokenizer st = new StringTokenizer(line);
      if(! st.hasMoreTokens()) return null;
      else return (CustomStatement)customStatements_.get(st.nextToken().toLowerCase());
    }
    
    void expandIf(
        String line,
        BufferedRandomAccessFile in,
        PrintWriter out,
        boolean visible)
        throws Exception {
        if (isErrorCase(line)) {
            handleErrorCase(in, out, visible);
        } else if (isNoResultCase(line)) {
            handleNoResultCase(line.trim(), in, out, visible);
        } else if (isPresenceCheck(line)) {
            handlePresenceCheck(line.trim(), in, out, visible);
        } else if (isFilePresenceCheck(line)) {
            handleFilePresenceCheck(line.trim(), in, out, visible);
        } else if (isEqual(line)) {
            checkEqual(line.trim(), in, out, visible);
        } else if (isCompareNumbers(line)) {
            compareNumbers(line.trim(), in, out, visible);
        } else if (isEval(line)) {
            expandEval(line, in, out, params_, visible);
        }
    } // method expandIf

    boolean startsWW(String base, String test) {
      //String s = base.trim().toUpperCase();
      //return s.startsWith(test + " ") || s.startsWith(test + "\t")
      //    || s.startsWith(test + "\n") || s.equals(test);

      int i=0;
      int l = base.length();
      while(i<l) {
        if(!Character.isWhitespace(base.charAt(i))) break;
        i++;
      }
      int t = test.length();
      if(!base.regionMatches(true,i,test,0,t)) return false;
      i+=t;
      if(i>=l) return true;
      char c = base.charAt(i);
      return (c == ' ' || c == '\t' || c == '\n');

    } // method startsWW

    boolean isComment(String s) {
        return s.trim().startsWith("$#");
    }
    boolean isAdmOnly(String s) {
        return startsWW(s, "$ADMINISTRATORONLY");
    }
    boolean isDebug(String s) {
        return startsWW(s, "$DEBUG");
    }
    boolean isAction(String s) {
        return startsWW(s, "$ACTION");
    }
    boolean isAssign(String s) {
        return startsWW(s, "$ASSIGN");
    }
    boolean isDefVar(String s) {
        return startsWW(s, "$DEFVAR");
    }
    boolean isSetCookie(String s) {
        return startsWW(s, "$SETCOOKIE");
    }
    boolean isCalc(String s) {
        return startsWW(s, "$ADD")
            || startsWW(s, "$SUB")
            || startsWW(s, "$MUL")
            || startsWW(s, "$DIV");
    }
    boolean isCheckNumber(String s) {
        return startsWW(s, "$CHECKNUMBER");
    }
    boolean isDefineSql(String s) {
        return startsWW(s, "$DEFINESQL");
    }
    boolean isMaterialize(String s) {
        return startsWW(s, "$MATERIALIZE");
    }
    boolean isSystem(String s) {
        return startsWW(s, "$SYSTEM");
    }
    boolean isAppendSql(String s) {
        return startsWW(s, "$APPENDSQL") || startsWW(s, "$APPENDSQL2");
    }
    boolean isAppend(String s) {
        return startsWW(s, "$APPEND");
    }
    boolean isExecSql(String s) {
        return startsWW(s, "$EXECSQL");
    }
    boolean isExpandSql(String s) {
      return startsWW(s, "$EXPANDSQL");
    }
    boolean isLoopBegin(String s) {
        return startsWW(s, "$LOOPONRESULTS")
            || startsWW(s, "$LOOPONENUM")
            || startsWW(s, "$LOOPONFILES");
    }
    boolean isLoopExit(String s) {
        return startsWW(s, "$EXITLOOP");
    }
    boolean isEndLoop(String s) {
        return startsWW(s, "$ENDLOOP");
    }
    boolean isErrorCase(String s) {
        return startsWW(s, "$IFERROR");
    }
    boolean isClearError(String s) {
        return startsWW(s, "$CLEARERROR");
    }
    boolean isNoResultCase(String s) {
        return startsWW(s, "$IFNORESULT");
    }
    boolean isCookie(String s) {
        return startsWW(s, "$COOKIE");
    }
    boolean isPresenceCheck(String s) {
        return startsWW(s, "$IFPRESENT");
    }
    boolean isFilePresenceCheck(String s) {
        return startsWW(s, "$IFFILEPRESENT");
    }
    boolean isEqual(String s) {
        return startsWW(s, "$IFEQUAL");
    }
    boolean isSwitch(String s) {
        return startsWW(s, "$SWITCH");
    }
    boolean isCompareNumbers(String s) {
        return startsWW(s, "$IFEE")
            || startsWW(s, "$IFLE")
            || startsWW(s, "$IFLT")
            || startsWW(s, "$IFGE")
            || startsWW(s, "$IFGT");
    }
    boolean isIf(String s) {
        return isErrorCase(s)
            || isNoResultCase(s)
            || isPresenceCheck(s)
            || isFilePresenceCheck(s)
            || isEqual(s)
            || isCompareNumbers(s)
            || isEval(s);
    }
    boolean isElse(String s) {
        return startsWW(s, "$ELSE");
    }
    boolean isEndIf(String s) {
        return startsWW(s, "$ENDIF");
    }
    boolean isExpandFile(String s) {
        return startsWW(s, "$EXPANDFILE");
    }
    boolean isInclude(String s) {
        return startsWW(s, "$INCLUDE");
    }
    boolean isCase(String s) {
        return startsWW(s, "$CASE");
    }
    boolean isDefaultCase(String s) {
        return startsWW(s, "$DEFAULT");
    }
    boolean isEndSwitch(String s) {
        return startsWW(s, "$ENDSWITCH");
    }
    boolean isEndCase(String s) {
        return startsWW(s, "$ENDCASE");
    }
    boolean isEval(String s) {
        return startsWW(s, "$IF");
    }
    boolean isForward(String s) {
      return startsWW(s, "$FORWARD");
    }
    // stored procedures
    boolean isDefineSqlProc(String s) {
        return startsWW(s, "$DEFINESQLPROC");
    }
    boolean isRegisterOutParameter(String s) {
        return startsWW(s, "$REGISTEROUTPARAMETER");
    }
    boolean isExecSqlProc(String s) {
        return startsWW(s, "$EXECSQLPROC");
    }

    // objects calls
    boolean isNewObject(String s) {
        return startsWW(s, "$NEWOBJECT");
    }
    boolean isInvoke(String s) {
        return startsWW(s, "$INVOKE");
    }
    boolean isCloseObject(String s) {
        return startsWW(s, "$CLOSEOBJECT");
    }
    
    void traceLine(String line) throws Exception {
        // Line syntax: $Debug text
        if (line.length() < 8)
            return;
        String traces = line.trim().substring(7).trim();
        logger_.fatal("$Debug: "+traces);
        logger_.fatal("$Debug with expand: "+expandString(traces));
    } // method traceLine

    void checkAdmLogin() throws IOException {
        if (!session_.isAdministrator())
            throw new IOException("ACCESS DENIED: Accessing this page requires ADMINISTRATOR rights");
    } // method checkAdmLogin

    /**
     * Assign a variable's value
     * Used for Defvar and Assign statements
     */
    void assignParam(String line) throws IOException, SQLException {
        // Line syntax: $Assign name val	or	$DefVar name val
        StringTokenizer st = new StringTokenizer(line);
        st.nextToken(); //Skip $Assign or $DefVar

        if (!st.hasMoreTokens())
            throw new IOException(syntaxError(line));
        String name = st.nextToken();

        if (!st.hasMoreTokens())
            throw new IOException(syntaxError(line));

        // 8 is the length of "$Assign " or "$DefVar ",
        // the "1" at the end skips one space
        int pos = 8 + line.substring(8).indexOf(name) + name.length() + 1;
        String value = line.substring(pos).trim();

        boolean expd = false; // value already expanded ? default false
        if (value.equals("\"\"")) {
            value = "";
            expd = true;
        } else if (value.equals("\" \"")) {
            value = " ";
            expd = true;
        }

        // Evaluate calculations (enclosed in parentheses)
        // eg. $Defvar sum ($var$ + 1)
        // Only numeric expressions are evaluated, boolean expressions are
        // considered as strings
        else if (value.startsWith("(") && value.endsWith(")")) {
            value = expandString(value);
            expd = true;
            ExpParser parser = new ExpParser(value, false);
            //no boolean exp. support
            ExpOperand operand = parser.parse();
            String w = parser.getWarning();
            if(w != null) logger_.debug(getPositionInfo(true) + w);
            if (operand != null
                && !operand.isStringValue()) { //do not eval strings
                ExpEval eval = new ExpEval();
                value = Double.toString(eval.eval(operand));
            }
        }

        if (name.startsWith("$") && name.endsWith("$"))
            name = expandString(name);

        setParameter(
            name,
            (expd ? value : expandString(value)),
            isAssign(line));

    } // method assignParam

    void appendParam(String line) throws IOException, SQLException {
        // Line syntax: $Append name val
        StringTokenizer st = new StringTokenizer(line);
        st.nextToken(); //Skip $Append

        if (!st.hasMoreTokens())
            throw new IOException(syntaxError(line));
        String name = st.nextToken();
        String prevval = (String) params_.get(name);
        if (prevval == null)
            throw new IOException("$Append: undefined variable " + name);

        if (!st.hasMoreTokens())
            throw new IOException(syntaxError(line));

        // 8 is the length of "$Append ", the "1" at the end skips one space
        int pos = 8 + line.substring(8).indexOf(name) + name.length() + 1;
        String value = line.substring(pos).trim();

        if (value.equals("\"\""))
            value = "";
        else if (value.equals("\" \""))
            value = " ";

        // Determine if it's a variable ($DefVar) or a parameter ($Assign)
        boolean isparm = (req_ != null && req_.getParameter(name) != null);
        setParameter(name, prevval + expandString(value), isparm);
    } // method appendParam

    void expandForward(String line) throws IOException {
      
      // Ignore forward if not applicable (eg. in ExpandFile or SendMail actions...)
      if(req_.getServletResponse() == null) return;
      
      // Line syntax: $Forward URL
      StringTokenizer st = new StringTokenizer(line);
      st.nextToken(); //Skip $Forward
      if (!st.hasMoreTokens())
        throw new IOException(syntaxError(line));
      javax.servlet.http.HttpServletRequest r = req_.getServletRequest();
      try {
        r.getRequestDispatcher(st.nextToken()).forward(r, req_.getServletResponse());
      } catch(javax.servlet.ServletException e) { throw new IOException(e.getMessage()); }
    }
    
    void setCookie(String line) throws IOException, SQLException {
        // Line syntax: $SetCookie name val
        StringTokenizer st = new StringTokenizer(line);
        st.nextToken(); //Skip $SetCookie

        if (!st.hasMoreTokens())
            throw new IOException(syntaxError(line));
        String name = st.nextToken();

        if (!st.hasMoreTokens())
            throw new IOException(syntaxError(line));

        // 11 is the length of "$SetCookie "
        // the "1" at the end skips one space
        int pos = 11 + line.substring(11).indexOf(name) + name.length() + 1;
        String value = line.substring(pos).trim();

        if (value.equals("\"\""))
            value = "";
        else if (value.equals("\" \""))
            value = " ";

        session_.getCookie().set(name, expandString(value));
    } // method setCookie

    void runAction(String line) throws Exception {
        logger_.debug("ESDynHtml:runAction " + line);

        // Line syntax: $Action actionClass

        StringTokenizer st = new StringTokenizer(line);
        st.nextToken(); //Skip $Action

        if (!st.hasMoreTokens())
            throw new IOException(syntaxError(line));

        String clazz = st.nextToken(); // The action class
        if (clazz.startsWith("$") && clazz.endsWith("$"))
            clazz =
                getVarValue(
                    clazz.substring(1, clazz.length() - 1),
                    params_,
                    false);

        String info = null;
        if (st.hasMoreTokens()) {
            // 8 is the length of "$Action ", the "1" at the end skips one space
            int pos = 8 + line.substring(8).indexOf(clazz) + clazz.length() + 1;
            info = line.substring(pos).trim();
            if (info.startsWith("$") && info.endsWith("$"))
                info =
                    getVarValue(
                        info.substring(1, info.length() - 1),
                        params_,
                        false);
        }

        ESAction act = (ESAction) Class.forName(clazz).newInstance();
        errorSet_ = !act.processRequest(session_, req_, this, info);

        setServletRequest(req_);
        // The parameters in the req. may have changed!
    } // method runAction

    /**
     * Define an SQL request
     * An SQL request may contain tags to expand (eg. "select * from $table$")
     * The tags will be expanded at run time.
     * @param line The $DefineSql statement
     */
    void defineSql(String line) throws Exception {

        // Line syntax: $DefineSql reqname sql
        // eg.	$DefineSql query1 select * from emp

        StringTokenizer st = new StringTokenizer(line);
        st.nextToken(); //Skip $DefineSql

        if (!st.hasMoreTokens())
            throw new IOException(syntaxError(line));
        String name = st.nextToken();

        if (!st.hasMoreTokens())
            throw new IOException(syntaxError(line));
        StringBuffer req = new StringBuffer(st.nextToken());
        while (st.hasMoreTokens())
            req.append(" " + st.nextToken());

        // Warning : DO NOT apply expandString() on the request
        // The expansion is done by execSql() at the last minute. This is
        // necessary so request can be parameterized, and changed according to
        // variable values.
        requests_.put(name, req.toString());
    } // method defineSql

    void appendSql(String line) throws Exception {

        boolean expand = false;

        // Line syntax: $AppendSql reqname clause
        // eg.	$AppendSql query1 AND Price>200

        StringTokenizer st = new StringTokenizer(line);
        String stmt = st.nextToken(); //Skip $AppendSql
        if (stmt.endsWith("2")) { //$AppendSql2 expands SQL strings immediately
            expand = true;
        }

        if (!st.hasMoreTokens())
            throw new IOException(syntaxError(line));

        String reqname = st.nextToken();
        String req = (String) requests_.get(reqname);
        if (req == null)
            throw new IOException(
                "$AppendSql: Undefined request: " + reqname + " line " + lnb_);

        if (!st.hasMoreTokens())
            throw new IOException(syntaxError(line));

        String clausestart = st.nextToken().toUpperCase();
        StringBuffer reqbuf = new StringBuffer(req.trim());

        String clause = line.substring((expand ? 12 : 11)
          + reqname.length()).trim();

        if (req.toUpperCase().endsWith("WHERE"))
            if (clausestart.equals("AND"))
                reqbuf.append(" " + clause.substring(4));
            else if (clausestart.equals("OR"))
                reqbuf.append(" " + clause.substring(3));
            else
                reqbuf.append(" " + clause);
        else
            reqbuf.append(" " + clause);

        if (expand) {
            // Expand string with displayUnresolved set to true
            // This causes unresolved variables to remain unchanged :
            // they will be resolved by execSql, so the user may change their value
            // between 2 executions of the SQL request
            requests_.put(
                reqname,
                expandString(reqbuf.toString(), params_, true));
        } else {
            requests_.put(reqname, reqbuf.toString());
        }
    } // method appendSql

    void execSql(String line) throws Exception {

        // Line syntax: $ExecSql reqname [datasource]
        // eg.	$ExecSql query1
        // eg.	$ExecSql query1 mysql1

        errorSet_ = false;

        StringTokenizer st = new StringTokenizer(line);
        st.nextToken(); //Skip $ExecSql

        if (!st.hasMoreTokens())
            throw new IOException(syntaxError(line));

        String reqname = st.nextToken();

        String req = (String) requests_.get(reqname);
        if (req == null)
            throw new Exception(
                "(execSql 2) Undefined request: " + req + " line " + lnb_);

        // Expand the SQL string (if some variables are present, they will be
        // expanded now - so the request is parameterized, it can be changed at
        // the last minute using a Defvar or Assign)
        String sql = expandString(req, params_);

        String datasource = null;
        if (st.hasMoreTokens()) {
            String d = st.nextToken();
            datasource = expandString(d, params_);
            if (datasource.compareTo("") == 0)
                datasource = null;
        }

        DataSet ds = null;
        Connection c = null;
        try {

          // Check if query starts with "[(*] [<whitespace>*] select"
          // (ignore case) - eg. select, (select, ( SELECT, etc...
          int i=0;
          int l = sql.length();
          while(i<l) {
            char cc = sql.charAt(i);
            if(! (cc == ' ' || cc == '\t' || cc == '(') ) break;
            i++;
          }

          //if (sql.toUpperCase().startsWith("SELECT")) {
          if(sql.regionMatches(true,i,"SELECT",0,6)) {

                logger_.debug(
                    "ESDynHtml:sql="
                        + sql
                        + " with datasource=<"
                        + datasource
                        + "> with session=<"+session_.toString()+"> started");
		session_.logRequest(datasource, sql);   

                c = session_.getConnection(datasource);
		
		Statement stat = null;
                try {
                  stat = c.createStatement(ResultSet.TYPE_SCROLL_SENSITIVE,
                    ResultSet.CONCUR_READ_ONLY);
                  if(stat == null) stat = c.createStatement();
                } catch(Exception e) {
                  if(stat != null) try { stat.close(); } catch(Exception ee) { }
                  stat = c.createStatement();
                }
                ds = new ESDBDataSet(
				    stat.executeQuery(sql),
				    session_,
				    c,
				    datasource, stat);
		
                logger_.debug(
			      "ESDynHtml:sql="
			      + sql
			      + " with datasource=<"
			      + datasource
			      + "> with session=<"+session_.toString()+"> ended");
                         
		addDataSet(reqname, ds);
          } else {
              logger_.debug("ESDynHtml:sql=" + sql);
              session_.updateDB(datasource, sql);
          }

        } catch (Exception e) {
            logError("sql=[" + sql + "] " + e);
            e.printStackTrace(System.err);
            try {
				if (ds != null) {
					datasets_.remove(reqname);
					ds.close();
				} else if (c != null)
					session_.closeConnection(c, datasource);
			} catch (Exception ee) {
			}
			errorSet_ = true;
            
        }

    }
    
    void expandSql(String line) throws Exception {

      // Line syntax: $ExpandSql reqname varname
      // eg.  $ExpandSql query1 sql

      StringTokenizer st = new StringTokenizer(line);
      st.nextToken(); //Skip $ExpandSql

      if (!st.hasMoreTokens())
          throw new IOException(syntaxError(line));

      String reqname = st.nextToken();
      
      if (!st.hasMoreTokens())
        throw new IOException(syntaxError(line));
      String varname = st.nextToken();
      
      String req = (String) requests_.get(reqname);
      if (req == null)
          throw new Exception(
              "(expandSql) Undefined request: " + req + " line " + lnb_);

      // Expand the SQL string into the specified variable
      // If some variables are present in the SQL string, they will be
      // expanded now
      setParameter(
          varname,
          expandString(req, params_),
          false);
    }

    boolean expandCase(
        String ln,
        BufferedRandomAccessFile in,
        PrintWriter out,
        String sw,
        boolean defaultcase)
        throws Exception {
        boolean aboolean = false;
        StringTokenizer st = new StringTokenizer(ln);
        st.nextToken().toUpperCase(); // Skip $Case or $Default
        if ((!st.hasMoreTokens()) && (!defaultcase))
            throw new IOException(syntaxError(ln));

        String cas = null;
        if (!defaultcase)
            cas = st.nextToken();
        String line;
        boolean endCase = false;
        String var = getVarValue(sw, params_, false);

        if ((cas != null) && (cas.startsWith("$")) && (cas.endsWith("$")))
            cas =
                getVarValue(cas.substring(1, cas.length() - 1), params_, false);

        if ((defaultcase) || (var.compareTo(cas) == 0)) {
            aboolean = true;
            while (!endCase && (line = in.readLine()) != null) {
                lnb_++;
                if(isSwitch(line)) expandSwitch(line, in, out, true); // Warning recursive call
                else if (isEndCase(line))
                    endCase = true;
                else
                    expandLine(line, in, out);
            }

        } else {
            while (!endCase && (line = in.readLine()) != null) {
                lnb_++;
                if(isSwitch(line)) expandSwitch(line, in, out, false); // Warning recursive call
                else if (isEndCase(line))
                    endCase = true;
            }
        }
        return aboolean;
    } // method expandCase

    void expandSwitch(String ln, BufferedRandomAccessFile in, PrintWriter out, boolean visible)
        throws Exception {

        StringTokenizer st = new StringTokenizer(ln);
        st.nextToken().toUpperCase(); //Skip $Switch

        if (!st.hasMoreTokens())
            throw new IOException(syntaxError(ln));

        String swit = st.nextToken();
        String line;
        boolean endSwitch = false;
        boolean found = false;

        while (!endSwitch && (line = in.readLine()) != null) {
            lnb_++;

            if (visible && (isCase(line)) && (!found))
                found = expandCase(line, in, out, swit, false);
            else if (isEndSwitch(line))
                endSwitch = true;
            else if (visible && (isDefaultCase(line)) && (!found))
                expandCase(line, in, out, swit, true);
        }

    } // method expandSwitch

    void expandLoop(String ln, BufferedRandomAccessFile in, PrintWriter out)
        throws Exception {

        // Line syntax: $<loopOp> datasetname [loopvarname [indexlow [indexhigh]]]
        // loopOp ::= LoopOnResults | LoopOnEnum | LoopOnFiles
        // Examples:
        //  $LoopOnResults dataset x
        //  $LoopOnResults dataset x 3
        //  $LoopOnResults dataset x 3 5
        //  $LoopOnResults dataset x $low$ $high$

        StringTokenizer st = new StringTokenizer(ln);
        String loopop = st.nextToken().toUpperCase(); //Get $<loopOp>

        if (!st.hasMoreTokens())
            throw new IOException(syntaxError(ln));

        String dsname = st.nextToken();

        String elemname;
        if (st.hasMoreTokens())
            elemname = st.nextToken();
        else
            elemname = new String("tuple");

        int ind1 = -1;
        int ind2 = -1;

        try {
            if (st.hasMoreTokens())
                ind1 = Integer.parseInt(expandString(st.nextToken()));
            if (st.hasMoreTokens())
                ind2 = Integer.parseInt(expandString(st.nextToken()));
        } catch (NumberFormatException ne) {
            throw new Exception(
                "ESDynHtml:expandLoop loop boundaries should be integers - line="
                    + lnb_);
        }

        if (ind1 >= 0 && ind2 >= 0 && ind2 < ind1)
            throw new Exception(
                "ESDynHtml:expandLoop check syntax (i>=1, j>=i) line=" + lnb_);

        String line = null;

        int prevlevel = looplevel_;
        DataSet ds = null;
        try {
			if (loopop.equals("$LOOPONENUM")) {
				addDataSet(dsname, new ESEnumDataSet(getVarValue(dsname,
						params_, false), ","));
			} else if (loopop.equals("$LOOPONFILES")) {
				String fnm = dsname;
				if (fnm.startsWith("$") && fnm.endsWith("$")) {
					fnm = getVarValue(fnm.substring(1, fnm.length() - 1),
							params_, false);
					dsname = fnm;
				}
				addDataSet(dsname, new ESFileDataSet(input_, fnm));
			}
			ds = (DataSet) datasets_.get(dsname);
			// Open the dataset here, and set boundaries if required:
			// This will allow to check if the dataset is empty, to skip the
			// loop
			if (ds != null) {
				ds.open();
				if (ind1 >= 0)
					ds.setBoundaries(ind1, ind2);
			}
			// If dataset is undefined or empty, skip the loop
			// Note the call to next(): if dataset is defined, it is now in 1st
			// position
			if (ds == null || !ds.next()) {
				if (looplevel_ == 0)
					setPriority(Thread.MIN_PRIORITY);
				looplevel_++; // for nested loops
				// skip the loop and return
				while (looplevel_ > prevlevel && (line = in.readLine()) != null) {
					lnb_++;
					if (isLoopBegin(line)) {
						if (looplevel_ == 0)
							setPriority(Thread.MIN_PRIORITY);
						looplevel_++;
					} else if (isEndLoop(line)) {
						looplevel_--;
						if (looplevel_ == 0)
							setPriority(Thread.NORM_PRIORITY);
						//TBD in.readLine();
					}
				}
				if (ds != null) {
				    if (ds.isTransient())
						datasets_.remove(dsname);
				    ds.close();
				}
				return;
			}
		} catch (Exception e) {
			if (ds != null) {
			    if (ds.isTransient())
				datasets_.remove(dsname);
			    ds.close();
			}
			throw e;
		}
        // Syntax (multi-line):
        // $<loopOp>
        // (lines to expand)
        // $EndLoop
        // When entering this function, the file pointer is positioned on the
        // $<loopOp> line.
        // The lines to expand will be expanded using the current ResultSet
		// first,
        // then using the parameters defined.
        // These lines should not contain instructions (like $DefineSql, etc...)

        // Save the current position
        loopstart_.put(dsname, new Long(in.getFilePointer()));
        int start_lnb = lnb_;

        // Loop on DataSet
        // The $<loopOp>...$EndLoop section will be expanded for each tuple
        // in the DataSet

        int nbtrace = 0;

        try {
			do { // do ... while(ds.next()) - see above, there was a 1st next() call
				logger_.debug("LOOP ON DATA SET : " + ++nbtrace);
				//Go back to the loop begin mark (useless the 1st time)
				Long fp = (Long) loopstart_.get(dsname);
				in.seek(fp.longValue());
				lnb_ = start_lnb;
				if (looplevel_ == 0)
					setPriority(Thread.MIN_PRIORITY);
				looplevel_++; // for nested loops
				logger_.debug("ESDynHtml:expandLoop avant appel getTuple");
				dataobjects_.put(elemname, ds.getTuple());
				logger_.debug("ESDynHtml:expandLoop apres appel getTuple");
				try {
				logger_.debug("ESDynHtml:expandLoop dans le try");

					// Expand each line in the loop section
					while (looplevel_ > prevlevel
							&& (line = in.readLine()) != null) {
					    logger_.debug("ESDynHtml:expandLoop dans le loop");
						lnb_++;
						if (isLoopBegin(line))
							expandLoop(line, in, out);
						else if (isEndLoop(line)) {
							looplevel_--;
							if (looplevel_ == 0)
								setPriority(Thread.NORM_PRIORITY);
						} else if (isExpandFile(line))
							expandFile(line, out, false);
						else if (isInclude(line)) {
							expandFile(line, out, true);
						} else if (isLoopExit(line)) {
							// Exception-based mechanism for $ExitLoop
							// isLoopExit(line) is tested at different places
							// (eg. while expanding $If... statements)
							// When a $If statement meets a $ExitLoop that should be
							// executed, it throws the exception too.
							throw new ESLoopExitException();
						} else {
							// Note: expandLine() may expand some blocks like $If ...
							// statements.
							// Some of these blocks may trigger a $ExitLoop:
							// if so, an ESLoopExitException will be thrown by the block,
							// and will be properly handled below (see the catch block).
						    logger_.debug("ESDynHtml:expandLoop avant le expand");
							expandLine(line, in, out);
						    logger_.debug("ESDynHtml:expandLoop apres le expand");
						}
					} // while (each line in the loop section)
				} catch (ESLoopExitException ee) {
				    logger_.debug("ESDynHtml:expandLoop dans le catch ESLoopExitException");
					// Skip the end of the loop, then throw the exception further
					// to leave the outermost loop (on the DataSet) and perform
					// necessary cleanup, like DataSet.close().
					// We may arrive here from this loop, or from an underlying
					// block (like a $If... statement)
					while (looplevel_ > prevlevel
							&& (line = in.readLine()) != null) {
						lnb_++;
						if (isLoopBegin(line)) {
							looplevel_++;
						} else if (isEndLoop(line)) {
							looplevel_--;
						}
					}
					if (looplevel_ <= 0)
						setPriority(Thread.NORM_PRIORITY);
					throw new ESLoopExitException();
				}
				if (line == null)
					throw new Exception("(expandLoop 2) Unexpected EOF line "
							+ lnb_);
			} while (ds.next());
		} catch (ESLoopExitException e) {
			// Ignore - Exception-based mechanism to exit loops...
			// Now, we left the DataSet loop, and will perform all necessary cleanup
			// (like DataSet.close(), etc...)
		} finally {
			dataobjects_.remove(elemname);
			loopstart_.remove(dsname);
			if (ds != null) {

			    logger_.debug("ESDynHtml:expandLoop avant appel du close");
			    ds.close();
			    if (ds.isTransient())
				datasets_.remove(dsname);
			    logger_.debug("ESDynHtml:expandLoop apres appel du close");
			}
		}
	} // method expandLoop

	void handleNoResultCase(String ln, BufferedRandomAccessFile in, PrintWriter out,
			boolean visible) throws Exception {
		// Line syntax: $IfNoResult dsname
		StringTokenizer st = new StringTokenizer(ln);
		st.nextToken(); //Skip $IfNoResult
		if (!st.hasMoreTokens())
			throw new IOException(syntaxError(ln));
		int prevlevel = iflevel_;
		iflevel_++;
		String dsname = st.nextToken();
		DataSet ds = (DataSet) datasets_.get(dsname);
		// If there's no result, display lines
		logger_.debug ("ESDynHtml:handleNoResultCase");
		/*if (ds.dataAvailable())
		    logger_.debug ("ESDynHtml:handleNoResultCase ok available");
		else
		    logger_.debug ("ESDynHtml:handleNoResultCase not available");
		*/
		boolean displayline = (ds == null || !ds.dataAvailable());
		// Close data set if no result
		if (ds != null && displayline) {
			datasets_.remove(dsname);
			ds.close();
		}
		String line;
		while (iflevel_ > prevlevel && (line = in.readLine()) != null) {
			lnb_++;
			if (isIf(line))
				expandIf(line, in, out, visible && displayline);
			else if (isEndIf(line))
				iflevel_--;
			else if (isElse(line))
				displayline = !displayline;
			else if (isLoopExit(line) && visible && displayline)
				// If not in a loop, the Exception will cause an error
				// (otherwise, it will result in a proper ExitLoop !)
				throw new ESLoopExitException("Unexpected $ExitLoop line "
						+ lnb_);
			else if (visible && displayline)
				expandLine(line, in, out);
		}
	} // method handleNoResultCase
	void handleErrorCase(BufferedRandomAccessFile in, PrintWriter out, boolean visible)
			throws Exception {
		// If there's an error, display lines
		boolean displayline = errorSet_;
		int prevlevel = iflevel_;
		iflevel_++;
		String line;
		while (iflevel_ > prevlevel && (line = in.readLine()) != null) {
			lnb_++;
			if (isIf(line))
				expandIf(line, in, out, visible && displayline);
			else if (isEndIf(line))
				iflevel_--;
			else if (isElse(line))
				displayline = !displayline;
			else if (isLoopExit(line) && visible && displayline)
				// If not in a loop, the Exception will cause an error
				// (otherwise, it will result in a proper ExitLoop !)
				throw new ESLoopExitException("Unexpected $ExitLoop line "
						+ lnb_);
			else if (visible && displayline)
				expandLine(line, in, out);
		}
	} // method handleErrorCase
	void handleCookie(PrintWriter out)
			throws java.io.UnsupportedEncodingException {
		println(out, "<input type=\"hidden\" name=\"S_\" value=\""
				+ java.net.URLEncoder.encode(session_.getClientId(), encodingValue) + "\" />");
	} // method handleCookie
	void handlePresenceCheck(String cmd, BufferedRandomAccessFile in, PrintWriter out,
			boolean visible) throws Exception {
		// Line syntax: $IfPresent Param [Param]
		// eg.	$IfPresent CustomerName
		StringTokenizer st = new StringTokenizer(cmd);
		st.nextToken(); //Skip $IfPresent
		if (!st.hasMoreTokens())
			throw new IOException(syntaxError(cmd));
		int prevlevel = iflevel_;
		iflevel_++;
		boolean present = visible;
		if(visible) {
		  while (st.hasMoreTokens()) {
				String field = st.nextToken();
				if (evalVariable(field, params_) == null)
				  present = false;
		  }
		}
		// If parameter is present, display lines
		boolean displayline = present;
		String line;
		while (iflevel_ > prevlevel && (line = in.readLine()) != null) {
			lnb_++;
			if (isIf(line))
				expandIf(line, in, out, visible && displayline);
			else if (isEndIf(line))
				iflevel_--;
			else if (isElse(line))
				displayline = !displayline;
			else if (isLoopExit(line) && visible && displayline)
				// If not in a loop, the Exception will cause an error
				// (otherwise, it will result in a proper ExitLoop !)
				throw new ESLoopExitException("Unexpected $ExitLoop line "
						+ lnb_);
			else if (visible && displayline)
				expandLine(line, in, out);
		}
	} // method handlePresenceCheck
	void handleFilePresenceCheck(String cmd, BufferedRandomAccessFile in,
			PrintWriter out, boolean visible) throws Exception {
		// Line syntax: $IfFilePresent Param [Param]
		// eg.	$IfPresent /usr/local/images/img.gif
		StringTokenizer st = new StringTokenizer(cmd);
		st.nextToken(); //Skip $IfFilePresent
		if (!st.hasMoreTokens())
			throw new IOException(syntaxError(cmd));
		int prevlevel = iflevel_;
		iflevel_++;
		boolean present = visible;
		if(visible) {
		  while (st.hasMoreTokens()) {
		    String fnm = st.nextToken();
		    if (fnm.startsWith("$") && fnm.endsWith("$"))
		      fnm = getVarValue(fnm.substring(1, fnm.length() - 1), params_,
		          false);
		    File tmp = new File(fnm);
		    File f;
		    if (tmp.isAbsolute())
		      f = tmp;
		    else {
		      String p = input_.getParent();
		      if (p == null)
		        f = tmp;
		      else if (p.endsWith(File.separator))
		        f = new File(p + tmp);
		      else
		        f = new File(p + File.separatorChar + tmp);
		    }
		    if (!f.exists())
		      present = false;
		  }
		}
		// If parameter is present, display lines
		boolean displayline = present;
		String line;
		while (iflevel_ > prevlevel && (line = in.readLine()) != null) {
			lnb_++;
			if (isIf(line))
				expandIf(line, in, out, visible && displayline);
			else if (isEndIf(line))
				iflevel_--;
			else if (isElse(line))
				displayline = !displayline;
			else if (isLoopExit(line) && visible && displayline)
				// If not in a loop, the Exception will cause an error
				// (otherwise, it will result in a proper ExitLoop !)
				throw new ESLoopExitException("Unexpected $ExitLoop line "
						+ lnb_);
			else if (visible && displayline)
				expandLine(line, in, out);
		}
	} // method handleFilePresenceCheck
	void checkNumber(String cmd) throws Exception {
		// Line syntax: $CheckNumber Value
		StringTokenizer st = new StringTokenizer(cmd);
		st.nextToken(); //Skip operator
		if (!st.hasMoreTokens())
			throw new IOException(syntaxError(cmd));
		String val = st.nextToken();
		if (val.startsWith("$") && val.endsWith("$"))
			val = getVarValue(val.substring(1, val.length() - 1), params_,
					false);
		try {
			Double.valueOf(val).doubleValue();
		} catch (NumberFormatException e) {
			errorSet_ = true;
		}
	} // method checkNumber
	void doCalc(String cmd) throws Exception {
		// Line syntax: $op Param Value
		// op ::= Add | Sub | Mul | Div
		StringTokenizer st = new StringTokenizer(cmd);
		String op = st.nextToken(); //Get operator
		if (!st.hasMoreTokens())
			throw new IOException(syntaxError(cmd));
		String varname = st.nextToken();
		if (!st.hasMoreTokens())
			throw new IOException(syntaxError(cmd));
		String val = st.nextToken();
		if (val.startsWith("$") && val.endsWith("$"))
			val = getVarValue(val.substring(1, val.length() - 1), params_,
					false);
		String varval = getVarValue(varname, params_, false);
		double d1, d2;
		try {
			d1 = Double.valueOf(varval).doubleValue();
			d2 = Double.valueOf(val).doubleValue();
		} catch (NumberFormatException nf) {
			logError(cmd + ": " + nf);
			throw (nf);
		}
		if (op.equalsIgnoreCase("$Add"))
			d1 += d2;
		else if (op.equalsIgnoreCase("$Sub"))
			d1 -= d2;
		else if (op.equalsIgnoreCase("$Mul"))
			d1 *= d2;
		else if (op.equalsIgnoreCase("$Div"))
			d1 /= d2;
		else
			throw new IOException(syntaxError(cmd));
		setParameter(varname, Double.toString(d1), true);
	} // method doCalc
	
	void expandEval(String ln, BufferedRandomAccessFile in, PrintWriter out,
			Hashtable params, boolean visible) throws Exception {
		String line = ln.trim().substring(3); // Skip $If
		String expline = expandString(line, params);
		boolean displayline = false;

		if(visible) {
		  ExpParser parser = new ExpParser(expline);
		  ExpEval eval = new ExpEval();
		  displayline = (eval.eval(parser.parse()) != 0);
                  String w = parser.getWarning();
                  if(w != null) logger_.debug(getPositionInfo(true) + w);
		}

		int prevlevel = iflevel_;
		iflevel_++;
		while (iflevel_ > prevlevel && (line = in.readLine()) != null) {
			lnb_++;
			if (isIf(line))
				expandIf(line, in, out, visible && displayline);
			else if (isEndIf(line))
				iflevel_--;
			else if (isElse(line))
				displayline = !displayline;
			else if (isLoopExit(line) && visible && displayline)
				// If not in a loop, the Exception will cause an error
				// (otherwise, it will result in a proper ExitLoop !)
				throw new ESLoopExitException("Unexpected $ExitLoop line "
						+ lnb_);
			else if (visible && displayline)
				expandLine(line, in, out);
		}
	} // method expandEval
	
	void compareNumbers(String cmd, BufferedRandomAccessFile in, PrintWriter out,
			boolean visible) throws Exception {
		// Line syntax: $op Param Value
		// op ::= IfEe | IfGt | IfGe | IfLt | IfLe
		StringTokenizer st = new StringTokenizer(cmd);
		String op = st.nextToken(); //Get operator
		if (!st.hasMoreTokens())
			throw new IOException(syntaxError(cmd));
		boolean displayline = false;
		if (visible) {
			String varname = st.nextToken();
			if (!st.hasMoreTokens())
				throw new IOException(syntaxError(cmd));
			String val = st.nextToken();
			if (val.startsWith("$") && val.endsWith("$"))
				val = getVarValue(val.substring(1, val.length() - 1), params_,
						false);
			String varval = getVarValue(varname, params_, false);
			double d1, d2;
			try {
				varval = varval.replace(',', '.');
				val = val.replace(',', '.');
				d1 = Double.valueOf(varval).doubleValue();
				d2 = Double.valueOf(val).doubleValue();
			} catch (NumberFormatException nf) {
				logError(cmd + ": " + nf);
				throw (nf);
			}
			if (op.equalsIgnoreCase("$IfEe"))
				displayline = d1 == d2;
			else if (op.equalsIgnoreCase("$IfLe"))
				displayline = d1 <= d2;
			else if (op.equalsIgnoreCase("$IfLt"))
				displayline = d1 < d2;
			else if (op.equalsIgnoreCase("$IfGe"))
				displayline = d1 >= d2;
			else if (op.equalsIgnoreCase("$IfGt"))
				displayline = d1 > d2;
			else
				throw new IOException(syntaxError(cmd));
		}
		int prevlevel = iflevel_;
		iflevel_++;
		String line;
		while (iflevel_ > prevlevel && (line = in.readLine()) != null) {
			lnb_++;
			if (isIf(line))
				expandIf(line, in, out, visible && displayline);
			else if (isEndIf(line))
				iflevel_--;
			else if (isElse(line))
				displayline = !displayline;
			else if (isLoopExit(line) && visible && displayline)
				// If not in a loop, the Exception will cause an error
				// (otherwise, it will result in a proper ExitLoop !)
				throw new ESLoopExitException("Unexpected $ExitLoop line "
						+ lnb_);
			else if (visible && displayline)
				expandLine(line, in, out);
		}
	} // method compareNumbers
	
	void checkEqual(String cmd, BufferedRandomAccessFile in, PrintWriter out,
			boolean visible) throws Exception {
		// Line syntax: $IfEqual Param Value
		// eg.	$IfEqual City New York
		StringTokenizer st = new StringTokenizer(cmd);
		st.nextToken(); //Skip $IfEqual
		if (!st.hasMoreTokens())
			throw new IOException(syntaxError(cmd));
		String varname = st.nextToken();
		if (!st.hasMoreTokens())
			throw new IOException(syntaxError(cmd));
		boolean displayline = visible;
		
		if(visible) {
		  String val = cmd.substring(9 + varname.length()).trim();
		  // If equality, display lines
		  if (val.equals("\"\"") || val.length() < 1)
		    val = "";
		  else if (val.startsWith("$") && val.endsWith("$"))
		    val = getVarValue(val.substring(1, val.length() - 1), params_,
		        false);
		  String val1 = getVarValue(varname, params_, false);
		  logger_.debug("IfEqual: compare [" + val1 + "] [" + val + "]");
		  displayline = val1.equals(val);
		}
		
		int prevlevel = iflevel_;
		iflevel_++;
		String line;
		while (iflevel_ > prevlevel && (line = in.readLine()) != null) {
			lnb_++;
			if (isIf(line))
				expandIf(line, in, out, visible && displayline);
			else if (isEndIf(line))
				iflevel_--;
			else if (isElse(line))
				displayline = !displayline;
			else if (isLoopExit(line) && visible && displayline)
				// If not in a loop, the Exception will cause an error
				// (otherwise, it will result in a proper ExitLoop !)
				throw new ESLoopExitException("Unexpected $ExitLoop line "
						+ lnb_);
			else if (visible && displayline)
				expandLine(line, in, out);
		}
	} // method checkEqual

  void expandMaterialize(String line) throws Exception {
    // Line syntax: $Materialize sourceDS destDS [ [-]column ]
    // eg. $Materialize Employees Sortedlist lastname
    StringTokenizer st = new StringTokenizer(line);
    st.nextToken(); //Skip $Materialize
    if (!st.hasMoreTokens())
      throw new IOException(syntaxError(line));
    String dsn = st.nextToken();
    if (dsn.startsWith("$") && dsn.endsWith("$"))
      dsn = getVarValue(dsn.substring(1, dsn.length()-1), params_, false);
    if (!st.hasMoreTokens())
      throw new IOException(syntaxError(line));
    String dss = st.nextToken();
    if (dss.startsWith("$") && dss.endsWith("$"))
      dss = getVarValue(dss.substring(1, dss.length()-1), params_, false);
    String col = null;
    boolean asc = true;
    if (st.hasMoreTokens()) {
      col = st.nextToken();
      if(col.startsWith("-")) {
        asc = false;
        col = col.substring(1);
      }
      if (col.startsWith("$") && col.endsWith("$"))
        col = getVarValue(col.substring(1, col.length()-1), params_, false);
    }
    try {
      ESMaterializeDataSet ds = new ESMaterializeDataSet((DataSet)datasets_.get(dsn));
      if(col != null) ds.orderBy(col, asc);
      addDataSet(dss, ds);
    } catch(Exception e) {
      throw new IOException("line " + lnb_
        + ": Unknown Dataset or wrong Dataset state for " + dsn + ": " + e);
    }
  }

  void expandSystem(String line, PrintWriter out) throws Exception {
    // Line syntax: $System command
    // eg.	$System rm /tmp/file.out
    StringTokenizer st = new StringTokenizer(line);
    st.nextToken(); //Skip $System
    if (!st.hasMoreTokens())
      throw new IOException(syntaxError(line));
    String s = st.nextToken();
    if (s.startsWith("$") && s.endsWith("$"))
      s = getVarValue(s.substring(1, s.length() - 1), params_, false);
    StringBuffer b = new StringBuffer(s);
    while (st.hasMoreTokens()) {
      s = st.nextToken();
      if (s.startsWith("$") && s.endsWith("$"))
        s = getVarValue(s.substring(1, s.length() - 1), params_, false);
      b.append(" " + s);
    }
    Process p = Runtime.getRuntime().exec(CmdTokenizer.tokenize(b.toString()));
    //Process p = Runtime.getRuntime().exec(b.toString());
    InputStream inp = p.getInputStream();
    byte buf[] = new byte[512];
    int n;
    while ((n = inp.read(buf)) >= 0)
      if (n > 0)
        writeString(out, new String(buf, 0, n));
  }
	
	void expandDefineSqlProc(String line) throws Exception {
		// Line syntax: $DefineSqlProc ProcedureName ProcedureCall [datasource]
		// eg.	$DefineSqlProc myprocedure {call aproc(?)} aDatasource
		StringTokenizer st = new StringTokenizer(line);
		String start = st.nextToken(); //Skip $DefineSqlProc
		String sql = expandString(line.substring(line.indexOf(start)
				+ start.length(), line.length()), params_);
		st = new StringTokenizer(sql);
		if (!st.hasMoreTokens())
			throw new IOException(syntaxError(sql));
		// next token is the name of the stored procedure
		String procName = st.nextToken();
		// get the call procedure
		int endbrak = sql.lastIndexOf("}");
		logger_.debug("line='" + line + "' procedure name='" + procName
				+ "', sql='" + sql + "'");
		String procedure = sql.substring(
				sql.indexOf(procName) + procName.length(), endbrak + 1).trim();
		String adatasource = null;
		logger_.debug("procedure name='" + procName + "', procedure='"
				+ procedure + "' line='" + line + "' endbrak='" + endbrak
				+ "' sql.length='" + sql.length() + "'");
		if (endbrak + 1 < sql.length())
			adatasource = sql.substring(endbrak + 1, sql.length()).trim();
		logger_.debug("procedure name='" + procName + "', procedure='"
				+ procedure + "', datasource='" + adatasource + "'");
		// create a CallableStatement object with a jdbc connection
		CallableStatement callable = null;
		ESCallable escall = null;
	
		Connection con = session_.getConnection(adatasource);
		session_.logRequest(adatasource, procedure);
		callable = con.prepareCall(procedure);
		// create an internal openeas object to manage the CallableStatement,
		// the out parameter, and the results
		escall = new ESCallable(callable, adatasource, con, session_);
		
		// adds this object to a list of procedure
		logger_.debug("create a procedure='" + procName + "'");
		procedures_.put(procName, escall);
	}

	void expandRegisterOutParameter(String line) throws Exception {
		// Line syntax: $RegisterOutParameter ProcedureName ParameterIndex ParameterType
		// eg.	$RegisterOutPamater myprocedure 1 VARCHAR
		StringTokenizer st = new StringTokenizer(line);
		st.nextToken(); //Skip $RegisterOutParameter
		if (!st.hasMoreTokens())
			throw new IOException(syntaxError(line));
		// get the name of the procedure
		String procName = st.nextToken();
		// if the procedure name is a variable
		if (procName.startsWith("$") && procName.endsWith("$"))
			procName = getVarValue(
					procName.substring(1, procName.length() - 1), params_,
					false);
		//procName += session_.getClientId();
		if (!st.hasMoreTokens())
			throw new IOException(syntaxError(line));
		// get the index of the parameter
		String number = st.nextToken();
		if (!st.hasMoreTokens())
			throw new IOException(syntaxError(line));
		// get the type of the parameter
		String varType = st.nextToken();
		// get the ESCallable object stored previously
		ESCallable callable = (ESCallable) (procedures_.get(procName));
		if (callable == null)
			throw new Exception(
					"Try to define out parameter without defining the procedure itself");
		// convert the parameter type into an 'int' to give it to the CallableStatement
		int typeNumber = string2Type(varType,
				(callable.getDatasource() == null) ? null : session_
						.getAppConfiguration().getName()
						+ "__" + callable.getDatasource());
		logger_.debug("procedure name='" + procName + "', number='" + number
				+ "', var type='" + varType + "' typeNumber='" + typeNumber
				+ "'");
		if (typeNumber == 999)
			throw new Exception("The type '" + varType
					+ "' is unknow, please see java.sql.Types");
		// register the given out parameter with the index of the parameter
		// and the integer parameter type
		callable.registerOutParameter(Integer.parseInt(number), typeNumber);
	}
	void expandExecSqlProc(String line) throws Exception {
		// Line syntax: $ExecSqlProc ProcedureName
		// eg.	$ExecSqlProc myprocedure
		StringTokenizer st = new StringTokenizer(line);
		st.nextToken(); //Skip $ExecSqlProc
		if (!st.hasMoreTokens())
			throw new IOException(syntaxError(line));
		// get the name of the procedure
		String procName = st.nextToken();
		// if the procedure name is a variable
		if (procName.startsWith("$") && procName.endsWith("$"))
			procName = getVarValue(
					procName.substring(1, procName.length() - 1), params_,
					false);
		//String orig = procName;
		//procName += session_.getClientId();
		// get the ESCallable object
		ESCallable callable = (ESCallable) (procedures_.get(procName));
		if (callable == null)
			throw new Exception(
					"Try to execute without defining the procedure itself");
		logger_.debug("procedure name='" + procName + "' toString='"
				+ callable.toString() + "'");
		try {
			callable.getCallable().execute();
			Hashtable params = callable.getParams();
			for (Enumeration enu = params.keys(); enu.hasMoreElements();) {
				String key = (String) (enu.nextElement());
				params.remove(key);
				String str = callable.getCallable().getString(
						Integer.parseInt(key));
				Object resultSet = callable.getCallable().getObject(
						Integer.parseInt(key));
				logger_.debug("key='" + key + "' str='" + str + "'");
				if (str != null)
					params.put(key, str);
				else if (resultSet == null)
					params.put(key, "");
				else {
					DataSet ds = new ESDBDataSet((ResultSet) (resultSet),
							session_, callable.getConnection(), callable.getDatasource(), null);
					logger_.debug("build a ESDBDataSet object with procname='"
							+ procName + key + "'");
					addDataSet(procName + key, ds);
				}
			}
			addDataObject(procName, callable);
		} catch (Exception e) {
			errorSet_ = true;
		}
	}
	/**
	 * create a new Java object
	 * @param line the line which contains the $NewObject keyword
	 * @throws Exception
	 */
	void expandNewObject(String line) throws Exception {
	  if (uObjects_ == null)
	    uObjects_ = new Hashtable();
	  StringTokenizer st = new StringTokenizer(line);
	  st.nextToken(); //Skip $NewObject
	  logger_.debug("expandNewObject line=<" + line + ">");
	  if (!st.hasMoreTokens())
	    throw new IOException(syntaxError(line));

	  // the following work is done to expand variable inside the list of parameters
	  // this works starts here ...

	  String persist = st.nextToken();
	  String rObj = null;
	  boolean isPersist = false;
	  if (persist.equalsIgnoreCase("isPersist")) {
	    // we must create a new Persistant UserObject
	    isPersist = true;
	    rObj = st.nextToken();
	  } else
	    rObj = persist;

	  //String rObj = st.nextToken();

	  logger_.debug("expandNewObject rObj=<" + rObj + ">");
	  String c = line.substring(line.indexOf(rObj) + rObj.length(),
	      line.length()).trim();
	  logger_.debug("expandNewObject c=<" + c + ">");
	  c = c.substring(1, c.length() - 1);
	  logger_.debug("expandNewObject c=<" + c + ">");
	  String paramsList = c.substring(c.indexOf('(') + 1, c.length() - 1);
	  logger_.debug("expandNewObject paramsList=<" + paramsList + ">");
	  if (paramsList.length() != 0)
	    line = line.substring(0, line.indexOf(paramsList))
	    + expandString(paramsList) + ")$";
	  logger_.debug("expandNewObject line=<" + line + ">");
	  // ... and this works ends here
	  UserObject uObj = null;
	  try {
	    uObj = new UserObject(line, lnb_, uObjects_, session_.getUserObjects(), logger_);
	  } catch (Exception e) {
	    // in case, we have an error when we try to call the constructor of the object
	    e.printStackTrace(System.err);
	    errorSet_ = true;
	    logger_.error("error ... exception");
	  }
	  if (uObj != null)
	    if (uObj.getName() == null)
	      errorSet_ = true;
	    else {
	      if (isPersist) {
	        //uObjects_.put(uObj.getName(), uObj);
	        session_.addUserObject(uObj.getName(), uObj);
	      }       
	      else
	        uObjects_.put(uObj.getName(), uObj);
	    }
	}
        
  /**
   * invokes a method on a specific object, previously created
   * @param line the line which contains $Invoke .....
   * @throws Exception
   */
  void expandInvoke(String line) throws Exception {
    if (uObjects_ == null)
      uObjects_ = new Hashtable();
    StringTokenizer st = new StringTokenizer(line);
    st.nextToken(); //Skip $Invoke
    if (!st.hasMoreTokens())
      throw new IOException(syntaxError(line));
      // the name of the object which will contains the result of the method

    String persist = st.nextToken();
    boolean isPersist = false;
    String objName = null;
                
    if (persist.equalsIgnoreCase("isPersist"))
      isPersist = true;
    else 
      objName = persist;

    String c = line.substring(line.indexOf(objName) + objName.length(),
      line.length()).trim();
    c = c.substring(1, c.length() - 1);
    String saveC = c;
    c = expandString(c);

    // the name of the object on which the method is used (not so english)
    int opar = c.indexOf('(');
    int dotp = c.substring(0, opar).lastIndexOf('.');
    String initObj = c.substring(0, dotp);

    UserObject uObj = null;

    if (uObjects_ != null)
      uObj = (UserObject) uObjects_.get(initObj);
    if (uObj == null)
      uObj = (UserObject) session_.getUserObject(initObj);

    if (uObj == null) // Supposed to be a static call...
      uObj = new UserObject(null, initObj, logger_);

    Object retObj = null;
    try {
      // invoke the method on the object, and return the result into retObj
      logger_.debug("expandInvoke c=<"+c+">");
      logger_.debug("expandInvoke saveC=<"+saveC+">");
      retObj = uObj.invoke(line.substring(0,line.indexOf(saveC))+c+"$",
       lnb_, uObjects_, session_.getUserObjects());
      logger_.debug("Create now a new UserObject");
      UserObject uoTemp = new UserObject(retObj, objName, logger_);
      logger_.debug("new UserObject has been created");
      if (uoTemp == null)
        logger_.debug("... and new UserObject is null");
      if (retObj == null)
        logger_.debug("... retObj is null");
      if (isPersist) {
        //uObjects_.put(objName, uoTemp);
        session_.addUserObject(objName, uoTemp);
      }
      else
        uObjects_.put(objName, uoTemp);
      logger_.debug("added object (uoTemp) into hashtable");
    } catch (Exception e) {
      errorSet_ = true;
      e.printStackTrace();
      logger_.error("INVOKE exception during method call, please verify syntax: "+ e);
      }
    // we use 2 types of objects, ESDataSet objects (then, we can loop on them), and
    // common objects (other), on which we only do a toString() call
    if (retObj == null)
      setParameter(objName, "null", isAssign(line));
                        
    if (retObj instanceof DataSet)
      addDataSet(objName, (DataSet) retObj);
    else
      setParameter(objName, (retObj != null ? retObj.toString() : "null"), isAssign(line));
    logger_.debug("End of INVOKE method");

  }
        
	/**
	 * Close an object
	 * @param line the line which contains $CloseObject .....
	 * @throws Exception
	 */
	void expandCloseObject(String line) throws Exception {
		StringTokenizer st = new StringTokenizer(line);
		st.nextToken(); //Skip $CloseObject     
		if (!st.hasMoreTokens())
			throw new IOException(syntaxError(line));
		// the name of the object which will contains the result of the method
                
                String objName = st.nextToken();
                
                if (session_.getUserObject(objName) == null) {
			errorSet_ = true;
			logger_.error("object (" + objName + ") does not exist");
		} else {
			session_.closeUserObject(objName);
		}
	}
        
	/**
	 * returns an integer which is the integer representation of a sql type
	 * or a specific 'sql' type (see OracleTypes for example).
	 * @param in    the name of the type (String representation)
	 * @return      an integer representation of the type
	 */
	int string2Type(String in, String adatasource) {
		// from java.sql.Types
		if ("BIT".compareTo(in) == 0)
			return Types.BIT;
		if ("TINYINT".compareTo(in) == 0)
			return Types.TINYINT;
		if ("SMALLINT".compareTo(in) == 0)
			return Types.SMALLINT;
		if ("INTEGER".compareTo(in) == 0)
			return Types.INTEGER;
		if ("BIGINT".compareTo(in) == 0)
			return Types.BIGINT;
		if ("FLOAT".compareTo(in) == 0)
			return Types.FLOAT;
		if ("REAL".compareTo(in) == 0)
			return Types.REAL;
		if ("DOUBLE".compareTo(in) == 0)
			return Types.DOUBLE;
		if ("NUMERIC".compareTo(in) == 0)
			return Types.NUMERIC;
		if ("DECIMAL".compareTo(in) == 0)
			return Types.DECIMAL;
		if ("CHAR".compareTo(in) == 0)
			return Types.CHAR;
		if ("VARCHAR".compareTo(in) == 0)
			return Types.VARCHAR;
		if ("LONGVARCHAR".compareTo(in) == 0)
			return Types.LONGVARCHAR;
		if ("DATE".compareTo(in) == 0)
			return Types.DATE;
		if ("TIME".compareTo(in) == 0)
			return Types.TIME;
		if ("TIMESTAMP".compareTo(in) == 0)
			return Types.TIMESTAMP;
		if ("BINARY".compareTo(in) == 0)
			return Types.BINARY;
		if ("VARBINARY".compareTo(in) == 0)
			return Types.VARBINARY;
		if ("LONGVARBINARY".compareTo(in) == 0)
			return Types.LONGVARBINARY;
		if ("NULL".compareTo(in) == 0)
			return Types.NULL;
		if ("OTHER".compareTo(in) == 0)
			return Types.OTHER;
		if ("JAVA_OBJECT".compareTo(in) == 0)
			return Types.JAVA_OBJECT;
		if ("DISTINCT".compareTo(in) == 0)
			return Types.DISTINCT;
		if ("STRUCT".compareTo(in) == 0)
			return Types.STRUCT;
		if ("ARRAY".compareTo(in) == 0)
			return Types.ARRAY;
		if ("BLOB".compareTo(in) == 0)
			return Types.BLOB;
		if ("CLOB".compareTo(in) == 0)
			return Types.CLOB;
		if ("REF".compareTo(in) == 0)
			return Types.REF;
		if ("DATALINK".compareTo(in) == 0)
			return Types.DATALINK;
		if ("BOOLEAN".compareTo(in) == 0)
			return Types.BOOLEAN;
		// in case the type has not been found in the java.sql.Types class,
		// we can find it in a specific given class
		Class types = null;
		try {
			if (adatasource != null) {
				Hashtable ds = session_.getAppConfiguration().getDatasources();
				logger_.debug("datasource='" + adatasource + "'");
				DatasourceProperty dsprop = (DatasourceProperty) (ds
						.get(adatasource));
				types = Class.forName(dsprop.getResultSet());
			} else
				types = Class.forName(session_.getAppConfiguration()
						.getResultSet());
		} catch (ClassNotFoundException cnfe) {
			logger_.debug("the class does not exist ! '"
					+ session_.getAppConfiguration().getResultSet() + "'");
		}
		try {
			Field field = types.getField(in);
			return field.getInt(null);
		} catch (NoSuchFieldException nsfe) {
			logger_.debug("the field does not exist !");
		} catch (IllegalAccessException iae) {
			logger_.debug("Illegal access to int type !");
		}
		// in case the type has not been found, we return a stupid value!
		return 999;
	}
	void expandFile(String line, PrintWriter out, boolean include)
			throws Exception {
		// Line syntax: $ExpandFile File	or	$Include File
		// eg.	$ExpandFile file2.tmpl
		StringTokenizer st = new StringTokenizer(line);
		st.nextToken(); //Skip $ExpandFile	or	$Include
		if (!st.hasMoreTokens())
			throw new IOException(syntaxError(line));
		// File name can be a variable ($fnm$) or a string
		String fnm = st.nextToken();
		if (fnm.startsWith("$") && fnm.endsWith("$")) {
			fnm = getVarValue(fnm.substring(1, fnm.length() - 1), params_,
					false);
		}
		File tmp = new File(fnm);
		File f = null;
		if (tmp.isAbsolute())
			f = tmp;
		else {
			String p = input_.getParent();
			if (p == null)
				f = tmp;
			else if (p.endsWith(File.separator))
				f = new File(p + tmp);
			else
				f = new File(p + File.separatorChar + tmp);
		}
		// Do not enforce file name security check
		// ExpandFile/Include are supposed to be used safely on the server side
		// Passing the file name as an HTTP parameter is dangerous !
		// (it may allow an unwanted read access on some files).
		
		// ESDynHtml can be overridden, and if so, the subclass must be instanciated
		// (the subclass may extend the DynHtml syntax...)
		// Replaces : ESDynHtml d = new ESDynHtml(session_, f, false);
		Class parameterTypes[] = { Client.class, File.class, boolean.class };
		java.lang.reflect.Constructor esd = this.getClass().getConstructor(parameterTypes);
		Object parm[] = { (org.experlog.util.Client)session_, f, new Boolean(false) };
		ESDynHtml d = (ESDynHtml)esd.newInstance(parm);
		d.setInner(true); // inner template (embedded in another template)

		d.requests_ = requests_;
		d.req_ = req_;
		d.params_ = params_;
		d.datasets_ = datasets_;
		d.dataobjects_ = dataobjects_;
		d.procedures_ = procedures_;
		d.expand(out, false);
		if (include) { // Expand file as if statements were included in the current file
			requests_ = d.requests_;
			req_ = d.req_;
			params_ = d.params_;
			datasets_ = d.datasets_;
			dataobjects_ = d.dataobjects_;
			procedures_ = d.procedures_;
		}
	} // method expandFile
	public String expandString(String line) throws SQLException {
		return expandString(line, params_, displayUnresolved_);
	}
	public String expandString(String line, Hashtable params)
			throws SQLException {
		return expandString(line, params, displayUnresolved_);
	}
	public String expandString(String line, Hashtable params,
			boolean displayUnresolved) throws SQLException {
		// Look for a variable name, to be expanded.
		// A variable name starts with a $, and ends with a $.
		// Example: "$Address$"
		int pos = line.indexOf('$');
		if (pos < 0)
			return line;
		StringBuffer b = new StringBuffer(line.substring(0, pos));
		pos++;
		String tmp = line.substring(pos);
		int endvar = tmp.indexOf('$');
		if (endvar < 0)
			return line; // May be a line with a price in $ !
		String varname = tmp.substring(0, endvar);
		pos += varname.length() + 1; //skip the last $
		boolean cookie = varname.toUpperCase().equals("COOKIE");
		if (cookie) { // COOKIE is expanded into a list of variables
		  try {
		    b.append("S_="
		        + java.net.URLEncoder.encode(
			    session_.getClientId(), encodingValue));
		  } catch(Exception e) {
		    b.append("S_=" + session_.getClientId());
		  }
		} else
			// Not COOKIE
			b.append(getVarValue(varname, params, displayUnresolved));
		// BEWARE!!
		// Recursive call
		b.append(expandString(line.substring(pos), params, displayUnresolved));
		return b.toString();
	} // method expandString
	
	public String getVarValue(String varname) throws SQLException {
	  if (varname.startsWith("$") && varname.endsWith("$"))
	    return getVarValue(varname.substring(1, varname.length()-1), params_, false);
	  else return varname;
	}
	
	String getVarValue(String varname, Hashtable params,
			boolean displayUnresolved) throws SQLException {
		// The "$\$" escape sequence will be expanded to "$"
		// This is a hack to express a price in $....
		// For example, the following sentence:
		// "Item price: $10, plus $2 shipping cost"
		// would cause openeas to expand a variable called "10, plus "
		// In that case, use:
		// "Item price: $\$10, plus $\$2 shipping cost"
		if (varname.equals("\\"))
			return "$";
		String unresolved = varname;
		// If the variable name starts with "(UrlEncode)", the var name will be
		// what follows (UrlEncode), and the value will be URL-encoded.
		// Example: "$(UrlEncode)Address$" will be replaced by the URL-encoded
		// value of the "Address" variable (if "Address" is "Bourbon Street",
		// it will be expanded into "Bourbon+Street").
		boolean urlencode = varname.toUpperCase().startsWith("(URLENCODE)");
		boolean sqlencode = varname.toUpperCase().startsWith("(SQLENCODE");
		boolean htmlencode = varname.toUpperCase().startsWith("(HTMLENCODE)");
                boolean xmlencode = varname.toUpperCase().startsWith("(XMLENCODE)");
		boolean jsencode = varname.toUpperCase().startsWith("(JSENCODE)");
		String datasourcename = null;
		if (sqlencode && varname.charAt(10) == '.') {
			// we have a '.' for a datasource
			datasourcename = varname.substring(11, varname.indexOf(")"));
		}
		if (urlencode || sqlencode || htmlencode || jsencode || xmlencode)
			varname = varname.substring(varname.indexOf(")") + 1);
		String fmt = null;
		if (varname.startsWith("(")) {
			int cpar = varname.indexOf(')');
			if (cpar >= 0) {
				// A format is defined
				fmt = varname.substring(1, cpar);
				varname = varname.substring(cpar + 1);
			}
		}
		String val = evalVariable(varname, params);
		if (val == null) {
			if (displayUnresolved)
				val = new String("$" + unresolved + "$"); // No change
			else
				val = new String("");
		} else if (fmt != null) {
			String fval;
			if (fmt.equalsIgnoreCase("Eval")) {
				if (val.indexOf('$') < 0) {
					fval = evalVariable(val, params);
					if (fval == null)
						fval = "";
				} else
					fval = expandString(val, params);
			} else {
				fval = formatValue(fmt, val);
			}
			val = fval;
		}
		if (urlencode)
		  try {
			return java.net.URLEncoder.encode(val, encodingValue);
		  } catch(Exception e) {
		    return val;
		  }
		else if (sqlencode)
			return sqlEncode(val, datasourcename);
		else if (htmlencode)
			return htmlEncode(val);
                else if (xmlencode)
			return xmlEncode(val);
		else if (jsencode)
			return jsEncode(val);
		else
			return val;
	} // method getVarValue
	
	String evalVariable(String name, Hashtable params) throws SQLException {
		if (name.startsWith("%"))
			return name.substring(1); //directly a value
		String val = null;
		String objname = null;
		String varname = null;
		int pos;
		if ((pos = name.indexOf(':')) > 0) {
			objname = name.substring(0, pos);
			varname = name.substring(pos + 1);
		} else
			varname = name;
		if (objname != null) {
			DataObject t = (DataObject) dataobjects_.get(objname);
			if (t != null)
				val = t.get(varname);
		} else {
			if (params != null) {
				val = (String) params.get(varname);
			}
			if (val == null) {
				// If the variable is a tuple name, display for debug
				DataObject t = (DataObject) dataobjects_.get(varname);
				if (t != null)
					val = t.toString();
			}
		}
		if (val != null)
			return val.trim();
		else
			return null;
	} // method evalVariable
	/**
	 * Format a decimal, date or string value.
	 * Dates will be converted from long values.
	 */
	String formatValue(String fmt, String val) {
		String fval = val;
		if (fmt.equalsIgnoreCase("Lower"))
		  return val.toLowerCase();
		else if (fmt.equalsIgnoreCase("Upper"))
		  return val.toUpperCase();
		else if (fmt.equalsIgnoreCase("UcFirst")) // 1st character upper case
		  return (val.length() < 2 ?
		      val.toUpperCase() : Character.toUpperCase(val.charAt(0)) + val.substring(1));
		else if (fmt.equalsIgnoreCase("RemoveHtmlTags"))
		  return StringUtils.removeHtmlTags(val);
		else if (fmt.equalsIgnoreCase("QuoteMessage"))
		  return StringUtils.quoteMessage(val);
		else if (fmt.equalsIgnoreCase("NormalizeUrl"))
		  return StringUtils.normalizeUrl(val);
		else if (fmt.startsWith("DateFormat.")) {
			// A date: convert from long (seconds elapsed since Jan 1, 1970)
			try {
				String customFmt = null;
				int df = DateFormat.DEFAULT;
				if (fmt.equals("DateFormat.DEFAULT"))
					df = DateFormat.DEFAULT;
				else if (fmt.equals("DateFormat.SHORT"))
					df = DateFormat.SHORT;
				else if (fmt.equals("DateFormat.MEDIUM"))
					df = DateFormat.MEDIUM;
				else if (fmt.equals("DateFormat.LONG"))
					df = DateFormat.LONG;
				else if (fmt.equals("DateFormat.FULL"))
					df = DateFormat.FULL;
				else
					//A custom format: example, DateFormat.yyyy/MM/dd
					//The format syntax is defined by java.text.SimpleDateFormat
					customFmt = fmt.substring(11); //skip "DateFormat."
				// Remove potential trailing ".0" or ".00" from value 
				DecimalFormat frm = new DecimalFormat("0",
						new DecimalFormatSymbols(formatLocale_));
				fval = frm.format(Double.valueOf(val).doubleValue());
				if (customFmt == null) {
					//Standard format: use a DateFormat object
					DateFormat formatter = DateFormat.getDateInstance(df,
							formatLocale_);
					fval = formatter.format(new java.util.Date(Long.parseLong(fval)));
				} else {
					//Custom format
					if (customFmt.startsWith("$") && customFmt.endsWith("$"))
						customFmt = getVarValue(customFmt, params_, false);
					SimpleDateFormat formatter = new SimpleDateFormat(
							customFmt, formatLocale_);
					fval = formatter.format(new java.util.Date(Long
							.parseLong(fval)));
				}
			} catch (Exception e) {
				fval = val;
			}
		} else if (fmt.startsWith("DateVal")) {
			// fmt can be, for example, "DateVal.DAY+1"
			// For the syntax, see the ESDateAccess class
			if (fmt.length() > 8)
				return dateaccess_.get(fmt.substring(8), fToDate(val + ":00",
						"yyyy:MM:dd:hh"));
			else
				return Long.toString(fToDate(val + ":00", "yyyy:MM:dd:hh"));
		} else if (fmt.startsWith("String")) {
			if (fmt.length() < 8)
				return val;
			String strf = fmt.substring(7);
			if (strf.equals("length")) { // Length of a String
				return Integer.toString(val.length());
				// null was already filtered out
			} else if (strf.startsWith("indexof:")) { // substring lookup
				try {
					String subf = strf.substring(8);
					if (subf.startsWith("%"))
						subf = subf.substring(1);
					else
						subf = expandString("$" + subf + "$");
					return (subf.length() > 0 ? Integer.toString(val
							.indexOf(subf)) : "-1");
				} catch (SQLException se) {
					return "-2";
				}
			} else if (strf.startsWith("sub")) {
				String subf = strf.substring(3);
				StringTokenizer st = new StringTokenizer(subf, ",");
				int lo = -1, hi = -1;
				if (!st.hasMoreTokens())
					return val;
				try {
					lo = Integer.parseInt(st.nextToken());
				} catch (Exception e) {
					return val;
				}
				if (val.length() < lo)
					return "";
				if (!st.hasMoreTokens())
					return val.substring(lo);
				try {
					hi = Integer.parseInt(st.nextToken());
				} catch (Exception e) {
					return val.substring(lo);
				}
				if (val.length() < hi)
					return val.substring(lo);
				else
					return val.substring(lo, hi);
			} else if (strf.startsWith("padl")) {
				String subf = strf.substring(4);
				int len = 0;
				try {
					len = Integer.parseInt(subf);
				} catch (Exception e) {
					return val;
				}
				if (val.length() >= len)
					return val.substring(0, len);
				StringBuffer b = new StringBuffer();
				for (int i = val.length(); i < len; i++)
					b.append(" ");
				b.append(val);
				return b.toString();
			} else if (strf.startsWith("padr")) {
				String subf = strf.substring(4);
				int len = 0;
				try {
					len = Integer.parseInt(subf);
				} catch (Exception e) {
					return "padr" + subf;
				}
				if (val.length() >= len)
					return val.substring(0, len);
				StringBuffer b = new StringBuffer(val);
				for (int i = val.length(); i < len; i++)
					b.append(" ");
				return b.toString();
			} else if (strf.startsWith("padc")) {
				String subf = strf.substring(4);
				int len = 0;
				try {
					len = Integer.parseInt(subf);
				} catch (Exception e) {
					return val;
				}
				if (val.length() >= len)
					return val.substring(0, len);
				int mid = (len - val.length()) / 2;
				StringBuffer b = new StringBuffer();
				for (int i = 0; i < mid; i++)
					b.append(" ");
				b.append(val);
				for (int i = 0; i < mid; i++)
					b.append(" ");
				if ((len - val.length()) % 2 != 0)
					b.append(" ");
				return b.toString();
			} else if(strf.startsWith("replace") && strf.length() >= 11) {
			  // replace<separatorChar>rules
			  // eg. replace;a=b;c=d; will change a into b and c into d
			  String rules = strf.substring(8);
			  return StringUtils.encodeWithRules(val, rules, strf.substring(7,8));
			} else
				return val;
		} else {
			// Not a date: decimal value assumed
			try {
				if (fmt.equals("Ceil"))
					return Double.toString(Math.ceil(Double.valueOf(val)
							.doubleValue()));
				else if (fmt.equals("Floor"))
					return Double.toString(Math.floor(Double.valueOf(val)
							.doubleValue()));
				else if (fmt.equals("Round"))
					return Double.toString(Math.round(Double.valueOf(val)
							.doubleValue()));
				DecimalFormat formatter = new DecimalFormat(fmt,
						new DecimalFormatSymbols(formatLocale_));
				fval = formatter.format(Double.valueOf(val).doubleValue());
			} catch (Exception e) {
				fval = val;
			}
		}
		return fval;
	} // method formatValue

	/**
	 * Convert a string to a date, given a date format
	 */
	long fToDate(String date, String fmt) {
	  SimpleDateFormat formatter = new SimpleDateFormat(fmt);
	  ParsePosition pos = new ParsePosition(0);
	  java.util.Date dt = formatter.parse(date, pos);
	  return dt.getTime();
	} // method fToDate

	public String sqlEncode(String s, String datasourcename)
			throws SQLException {
		String sep = session_.getAppConfiguration().get("sqlencode.Separator");
		if (sep == null)
			sep = ";";
		//		String rules = session_.getShopConfiguration().get("sqlencode.Substitutions");
		String rules = session_.getAppConfiguration().getSqlEncodeSubst(
				datasourcename);
		// Rules syntax: c1=s1;c2=s2;...cn=sn (replace char1 with string1, etc...)
		// ; is the default separator (sqlencode.separator to change it)
		// default rules: replace ' with '', and newlines with spaces.
		if (rules == null)
			rules = "\'=\'\';\r=\" \";\n=\" \";";
		return StringUtils.encodeWithRules(s, rules, sep);
	} // method sqlEncode

	/**
	 * Value escaping for HTML strings (eg. in "hidden" tags):
	 * Replace " with &quot;, & with &amp;, > with &gt; and < with &lt;
	 */
	String htmlEncode(String s) {
		return StringUtils.encodeWithRules(
				s,
				"\"=&quot;|&=&amp;|<=&lt;|>=&gt;|=&eacute;|=&egrave;|=&agrave;|=&ugrave;|=&acirc;|=&ecirc;|=&icirc;|=&ocirc;|=&ucirc;|=&ccedil;|",
				"|");
	}
     
	String xmlEncode(String s) {
	  return StringUtils.encodeWithRules(
	      s,
	      "&=&amp;|<=&lt;|>=&gt;|%=&#37;|^=&#94;|\"=&quot;|'=&apos;|\n=&#10;|",
	  "|");
	}
        
	/**
	 * Value escaping for javascript strings:
	 * Replace " with \" and ' with \'
	 */
	String jsEncode(String s) {
		return StringUtils.encodeWithRules(s,
				"\'=\\\'|\"=\\\"|\n=\\n|\r=\\r|\\=\\\\|\t=\\t", "|");
	}

	private void setPriority(int pri) {
		try {
			Thread.currentThread().setPriority(pri);
		} catch (Exception e) {
			//ignore
		}
	} // method setPriority
	private void flush(PrintWriter out) {
		if (out != null)
			out.flush();
	}
	private void println(PrintWriter out, String text) {
		if (out != null)
			out.println(text);
	}
	private void writeString(PrintWriter out, String text) {
		if (out != null)
			out.write(text);
	}
	private String syntaxError(String msg) {
		return "Syntax error line " + lnb_ + ": " + msg;
	}
	// A test program
	public static void main(String args[]) {
		try {
			DynHtml exp = new ESDynHtml(null, new File("DynHtml.tmpl"));
			exp.setParameter("firstname", "Pierre-Yves", false);
			exp.setParameter("lastname", "Gibello", false);
			PrintWriter out = new PrintWriter(System.out);
			exp.expand(out);
			out.flush();
		} catch (Exception e) {
			System.err.println("ESDynHtml Exception " + e);
		}
	} // method main
} // class ESDynHtml
