/*
 * OpenEAS Experlog Application Server
 * Web application development and dynamic template scripting engine
 * Copyright (c) 2004 ExperLog SARL - http://www.experlog.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version. This program is distributed in
 * the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
 * the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have
 * received a copy of the GNU General Public License along with this program;
 * if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */


package org.experlog.openeas.api;

import org.experlog.db.ESDataSourceWrapper;
import org.experlog.util.ESCategory;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.Hashtable;

import org.apache.log4j.Category;

public class ConnectionManager {

    /**
     * structure which contains all datasources from all application
     * key: the name of the datasource
     * value: ESDataSourceWrapper object
     */
    private static Hashtable datasources = new Hashtable();
    private static Category logger = ESCategory.getCategory("DataSourceCenter", "openeas.db");

    /**
     * add a datasource to the list of known datasources
     * @param esdw an ESDataSourceWrapper object
     */
    public static void addDataSource(ESDataSourceWrapper esdw) {
        logger.debug("ESDataSourceCenter:add =<"+esdw.getDataSourceName()+">");
        datasources.put(esdw.getDataSourceName(), esdw);
        logger.debug("ESDataSourceCenter:add size=<"+datasources.size()+">");
    }

    /**
     * return a datasource with a datasource name
     * @param datasource the name of the datasource
     * @return  an ESDataSourceWrapper object which is represented by its name
     */
    private static ESDataSourceWrapper getDataSource(String datasource) {
        logger.debug("ESDataSourceCenter:getDataSource =<"+datasource+
                "> size=<"+datasources.size()+">");
        return (ESDataSourceWrapper)datasources.get(datasource);
    }

    /**
     * return a connection when the datasource name is given
     * @param datasource the name of the datasource
     * @return a connection object
     * @throws SQLException
     * @throws Exception
     */
    public static Connection getConnection(String datasource)
            throws SQLException, Exception {
        ESDataSourceWrapper es = getDataSource(datasource);
        if (es == null)
            logger.error("ESDataSourceCenter:getConnection datasource is null!");
        else
            return es.getConnection();
        return null;
    }

}
