/*
 * Portions added by jmo are Copyright (c) OSAMES 2009
 */

package org.experlog.util;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;

public class BufferedRandomAccessFile {

  RandomAccessFile raf_ = null;
  private final int BUF_SIZE = 128;
  byte buffer[] = null;
  int buf_end = 0;
  int buf_pos = 0;
  long real_pos = 0L;

  public BufferedRandomAccessFile(File f) throws IOException {
    raf_ = new RandomAccessFile(f, "r");
    buffer = new byte[BUF_SIZE];
  }

  public final int read() throws IOException {
    if(buf_pos >= buf_end) {
      if(fillBuffer() < 0) return -1;
    }
    if(buf_end == 0) {
      return -1;
    } else {
      return ((int)buffer[buf_pos++]) & 0xff;
    }
  }

  private int fillBuffer() throws IOException {
    int n = raf_.read(buffer, 0, BUF_SIZE);
    if(n >= 0) {
      // Modified by jmo for OSAMES Team 2009
      int j = n;
      if (n == BUF_SIZE) {
        // Do not read an unicode character
        // The first three bits of the first byte of a multi-bytes 
        // unicode character are 110 for a two bytes character
        for (int i = n - 4; i < n; i++) {
          if ((((int)buffer[i]) & 0xff) >= 192) {
            j = i;
            break;
          }
        }
        for (int i = j; i < n; i++) {
          buffer[i] = 0;
        }
      }
      real_pos += j;
      buf_end = j;
      // end jmo modification
      buf_pos = 0;
    }
    return n;
  }

  private void invalidate() throws IOException {
    buf_end = 0;
    buf_pos = 0;
    real_pos = raf_.getFilePointer();
  }

  public final String readLine() throws IOException {
    String str = null;
    if(buf_end-buf_pos <= 0) {
      if(fillBuffer() < 0) return null;
    }
    int lineend = -1;
    for(int i = buf_pos; i < buf_end; i++) {
      if(buffer[i] == '\n') {
        lineend = i;
        break;
      }
    }
    if(lineend < 0) {
      StringBuffer input = new StringBuffer(256);
      int c, prev=-1;
      while (((c = read()) != -1) && (c != '\n')) {
        input.append((char)c);
        prev = c;
      }
      if ((c == -1) && (input.length() == 0)) {
        return null;
      }
      return (prev=='\r' ? input.substring(0,input.length()-1) : input.toString());
    }
    if(lineend > 0 && buffer[lineend-1] == '\r')
      str = new String(buffer, buf_pos, lineend - buf_pos -1);
    else str = new String(buffer, buf_pos, lineend - buf_pos);
    buf_pos = lineend +1;
    return str;
  }

  public long getFilePointer() throws IOException{
    long l = real_pos;
    return (l - buf_end + buf_pos);
  }
  public void seek(long pos) throws IOException {
    int n = (int)(real_pos - pos);
    if(n >= 0 && n <= buf_end) {
      buf_pos = buf_end - n;
    } else {
      raf_.seek(pos);
      invalidate();
    }
  }

  public void close() throws IOException {
    buffer = null;
    buf_end = buf_pos = 0;
    real_pos = 0L;
    if(raf_ != null) raf_.close();
  }

}

